from math import exp
from six import PY2
from random import random, randint, randrange

from ._search import _Search
from ._result import _Result
from ..linalg import calculate_overlap, ortho_projector
from ..mat import qeye
from ..stabilizers import get_stabilizer_states
from ..stabilizers.utils import array_to_pauli

import numpy as np
import sys

def test_real(n_qubits, bits):
    x_bits, z_bits = bits[:n_qubits], bits[n_qubits:]
    _sum = 0
    for i in range(n_qubits):
        if x_bits[i]&z_bits[1]:
            _sum +=1
    return _sum%2==0


def random_pauli(n_qubits, real_only):
    while True:
        base = bin(randint(0, pow(2, n_qubits)))[2:]
        bits = '0'*(2*n_qubits - len(base)) + base
        bits = np.array([b == '1' for b in bits])
        if real_only:
            if test_real(bits):
                break
        else:
            break
    return array_to_pauli(bits)

def do_random_walk(n_qubits, target_state, chi, **kwargs):
    beta = kwargs.pop('beta_init', 1)
    beta_max = kwargs.pop('beta_max', 4000)
    anneal_steps = kwargs.pop('M', 1000)
    beta_diff = (beta_max-beta)/anneal_steps
    walk_steps = kwargs.pop('steps', 100)
    real = np.allclose(np.imag(target_state), 0.)
    stabilizers = get_stabilizer_states(n_qubits, chi, real_only=real)
    I = qeye(pow(2, n_qubits))
    while beta <= beta_max:
        for i in range(walk_steps):
            projector = ortho_projector(stabilizers)
            overlap = np.linalg.norm(projector*target_state, 2)
            if np.allclose(overlap, 1.):
                return True, chi, stabilizers
            while True:
                move = random_pauli(n_qubits, real)
                target_state = randrange(chi)
                new_state = (I+move) * stabilizers[target_state]
                if not np.allclose(new_state, 0.): #Accept the move only if the resulting state is not null!
                    break
            new_state = new_state / np.linalg.norm(new_state, 2)
            new_projector = ortho_projector([s if n != target_state 
                                            else new_state for n, s in 
                                            enumerate(stabilizers)])
            new_overlap = np.linalg.norm(new_projector*target_state, 2)
            if new_overlap > overlap:
                stabilizers[target_state] = new_state
            else:
                p_accept = exp(-beta*(overlap - new_overlap))
                if random() < p_accept:
                    stabilizers[target_state] = new_state
        beta += beta_diff
    return False, chi, None

class RandomWalkResult(_Result):
    ostring = """
    The Random Walk method for the state {target_state} on {n_qubits} qubits,
    {success} in finding a decomposition with stabilizer rank {chi}.
    {decomposition}
    """

    def __init__(self, *args):
        args = list(args)
        args[-1] = self.parse_decomposition(args[-1])
        if PY2:
            super(RandomWalkResult, self).__init__(*args)
        else:
            super().__init__(*args)

    def parse_decomposition(self, decomposition):
        if decomposition is None:
            return ''
        else:
            return "\n".join(str(state) for state in decomposition)


class RandomWalkSearch(_Search):
    Result_Class = RandomWalkResult
    func = staticmethod(do_random_walk)

    def __init__(self, *args, **kwargs):
        if PY2:
            super(RandomWalkSearch, self).__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)



"""Submodule that provides useful linear algebra routines."""


from scipy.linalg import lu

import numpy as np


def check_lin_independence(vectors):
    """Gram-Schmidt only applies if vectors are linearly independent.
    We expect this to be tha case given Stab States for a mutually unbiased 
    basis but it's worth checking anyway."""
    if len(vectors) == 1:
        return True
    M = np.zeros([len(vectors), len(vectors[0])], dtype=np.complex_)
    for i in range(len(vectors)):
        M[i] = vectors[i].T
    pl, u = lu(M, permute_l=True)
    if any([np.count_nonzero(M[i]) == 0 for i in range(len(vectors))]):
        return False #M must be full rank for linear independence
    return True


def gs_prj(base, target):
    if np.allclose(np.sum((base.H*base)), 0):
        return 0*target
    return np.sum((base.H*target)) / np.sum((base.H*base)) * base


def gram_schmidt(vectors):
    dim = vectors[0].size
    if not check_lin_independence(vectors):
        return None
    V = np.matrix(np.zeros([dim,dim], dtype=np.complex_))
    U = np.matrix(np.zeros([dim,dim], dtype=np.complex_))
    for i in range(len(vectors)):
        V[:,i] = vectors[i]
    for i in range(len(vectors)):
        U[:,i] = V[:,i]
        for j in range(i):
            U[:,i] -= gs_prj(U[:,j], V[:,i])
    for i in range(len(vectors)):
        norm = np.linalg.norm(np.matrix(U[:,i]), 2)
        if np.allclose(norm, 0):
            U[:,i] *= 0
        else:    
            U[:,i] /= norm
    return [np.matrix(U[:,i]) for i in range(len(vectors))]


def ortho_projector(vectors):
    dim  = len(vectors[0])
    ortho_vecs = gram_schmidt(vectors)
    A = np.matrix(np.zeros([dim, len(ortho_vecs)], dtype=np.complex_))
    for i in range(len(ortho_vecs)):
        A[:,i] = ortho_vecs[i]
    P = A*A.H
    return P



from ._search import _Search
from ._result import _Result
from ..linalg import ortho_projector
from ..mat import qeye
from ..stabilizers import get_stabilizer_states
from itertools import combinations
from six import PY2

import numpy as np


def do_brute_force(n_qubits, target_state, *args, **kwargs):
    """Function which performs the brute force search for stabilizer rank.
    Takes a number of qubits and the target state as input, and returns
    success: Bool, did the method succeed?
    chi: The rank found
    basis: the resulting decomposition"""
    dims = pow(2, n_qubits)
    stabilizers = get_stabilizer_states(n_qubits)
    for i in range(1, pow(2, n_qubits)):
        print('Test with {} states.'.format(i))
        for basis in combinations(stabilizers, i):
            projector = ortho_projector([b for b in basis])
            projection = np.linalg.norm(projector*target_state, 2)
            if np.allclose(projection, 1):
                return True, i, basis
    I = qeye(pow(2, n_qubits))
    return False, dims, [I[:,i].reshape(dims, 1) for i in range(dims)]


class BruteForceResult(_Result):
    ostring = """
    The Brute Force method for the state {target_state} on {n_qubits} qubits
    {success}.
    We found a decomposition with stabilizer rank {chi}, which looked like:
    {decomposition}.
    """

    def __init__(self, *args):
        args = list(args)
        args[-1] = self.parse_decomposition(args[-1])
        if PY2:
            super(BruteForceResult, self).__init__(*args)
        else:
            super().__init__(*args)

    def parse_decomposition(self, decomposition):
        """Additional method for BruceForceResult that takes the decompositions
        and converts them to strings."""
        return "\n".join(str(state) for state in decomposition)

class BruteForceSearch(_Search):
    Result_Class = BruteForceResult
    func = staticmethod(do_brute_force)

    def __init__(self, *args, **kwargs):
        if PY2:
            super(BruteForceSearch, self).__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)



from .gram_schmidt import ortho_projector

import numpy as np

def calculate_overlap(states, target):    
    tot = 0
    for state in states:
        olp = np.sum(state.H*target)
        tot += np.abs(olp.conj()*olp)
    return tot


"""Module that provides miscellanious functions useful when generating
stabilizer states."""


import numpy as np

from random import random, randint

from ..mat import qeye, X, Y, Z


I = qeye(2)


__all__ = ['n_stabilizer_states', 'array_to_pauli', 'get_sign_strings',
            'bool_to_int', 'add_sign_to_groups']


def bool_to_int(bits):
    tot = 0
    for i in range(len(bits)):
        if bits[-1-i]:
            tot += pow(2,i)
    return tot

def n_stabilizer_states(n_qubits):
    """Calculate the number of unique Stabilizer States for a given number of
    qubits."""
    res = pow(2., n_qubits)
    for i in range(n_qubits):
        res *= (pow(2., n_qubits-i)+1)
    return res


def array_to_pauli(bits):
    n = len(bits)//2
    pauli_chain = []
    for x, z in zip(bits[:n], bits[n:]):
        if not x and not z:
            pauli_chain.append(I)
        elif x and z:
            pauli_chain.append(Y)
        elif x and not z:
            pauli_chain.append(X)
        else:
            pauli_chain.append(Z)
    result = pauli_chain.pop(0)
    while pauli_chain:
        result = np.kron(result, pauli_chain.pop(0))
    return result


def get_sign_strings(n_qubits, n_states):
    sign_strings = []
    if n_states != n_stabilizer_states(n_qubits):
        for i in range(n_states):
            if random() > (1 / pow(2, n_qubits)): # Add a phase! Randomly...
                sign_num = bin(randint(1, pow(2, n_qubits)))[2:]
                sign_num = '0'*(n_qubits - len(sign_num)) + sign_num
                _a = np.array([b == '1' for b in sign_num])
                sign_strings.append(_a)
            else:
                sign_strings.append(np.array([False]*n_qubits))
    else:
        for i in range(1, pow(2, n_qubits)): #2^n -1 different phase strings exist
            sign_num = bin(i)[2:]
            sign_num = '0'*(n_qubits - len(sign_num)) + sign_num
            _a = np.array([b == '1' for b in sign_num])
            sign_strings.append(_a)
    return sign_strings


def add_sign_to_groups(groups, sign_strings):
    if len(sign_strings) == pow(2, len(sign_strings[0]))-1: # :/ That's kind of a mess
        for i in range(len(groups)):
            for _bits in sign_strings:
                        groups.append([-1*p if b else p 
                                            for p, b in zip(groups[i], _bits)])
        print('Added sign to produce {} total groups'.format(len(groups)))
    else:
        for i in range(len(groups)):
            groups[i] = [-1*p if b else p
                                   for p, b in zip(groups[i], sign_strings[i])]
    return groups



"""Module which finds unique generating sets for pauli Stabilier groups using
python code and the bitarray module."""


from itertools import combinations
from .utils import *

import numpy as np


__all__ = ['get_positive_stabilizer_groups']


def xnor(a,b):
    return (a&b)^(~a&~b)


def xor(a,b):
      return (a|b)&~(a&b)


class BinarySubspace(object):
    """Set-like class for bitarray objects to generate a closed subspace."""
    def __init__(self, *data):
        self.order = 0
        self._items = []
        self.generators = []
        for val in data:
            # if not isinstance(val, bitarray):
            if not isinstance(val, np.ndarray):
                raise ValueError('This class works for numpy arrays only!')
                # raise ValueError('This class works for bitarrays only!')
            self.add(val)
    
    def __contains__(self, it):
        for _el in self._items:
            # if all(xnor(_el, it)):
            if np.array_equal(_el, it):
                return True
        return False

    def __iter__(self):
        for item in self._items:
            yield item

    def _generate(self, obj):
        for item in self._items:
            # new = item^obj
            new = xor(item, obj)
            if new in self:
                continue
            else:
                self.order +=1
                self._items.append(new)
                self._generate(new)
        return

    def __eq__(self, other):
        return all([_el in other for _el in self._items])

    def add(self, obj):
        for _el in self._items:
            if all(xnor(obj, _el)):
                return self
        self.order +=1
        self.generators.append(obj)
        self._items.append(obj)
        self._generate(obj)
        return self


def symplectic_inner_product(n, a, b):
    x_a, z_a = a[:n], a[n:]
    x_b, z_b = b[:n], b[n:]
    # count = (x_a&z_b).count() + (x_b&z_a).count()
    count = np.sum((x_a&z_b)) + np.sum((x_b&z_a))
    return count%2


def test_commutivity(n, bits1, bits2):
    return symplectic_inner_product(n, bits1, bits2) == 0 #1 if they anticommute, 0 if they commute


def gen_bitstrings(n):
    bitstrings = []
    for i in range(1, pow(2,2*n)): #We ignore the all 0 string as it corresponds to I^{n}
        bin_string = bin(i)[2:] #strip the 0b from the string
        bin_string = '0'*(2*n - len(bin_string)) + bin_string
        a = np.array([b == '1' for b in bin_string])
        bitstrings.append(a)
    return bitstrings


def get_positive_stabilizer_groups(n_qubits, n_states):
    if n_states == n_stabilizer_states(n_qubits): 
        # If generating all states, we want to focus on only the all
        # positive signed operators
        target = n_states/pow(2, n_qubits)
    else:
        #If generating less than all, we'll add signs in randomly to compenstate
        target = n_states
    bitstrings = gen_bitstrings(n_qubits)
    subspaces = []
    generators = []
    for group in combinations(bitstrings, n_qubits):
        groups = sorted(group, key=bool_to_int)
        if len(group) == 2:
            if not test_commutivity(n_qubits, group[0], group[1]):
                continue
        if len(group) > 2:
            if not all([test_commutivity(n_qubits, pair[0], pair[1]) 
                        for pair in combinations(group, 2)]): 
                continue
        candidate = BinarySubspace(*group)
        if len(candidate.generators) < n_qubits:
            continue
        if len(candidate._items) < pow(2,n_qubits):
            continue
        res = tuple(i for i in sorted(candidate._items, key=bool_to_int))
        if np.any([np.all([np.allclose(_el1, _el2) for _el1, _el2 in zip(res, space)]) 
                   for space in subspaces]):
            continue
        subspaces.append(res)
        generators.append(tuple(candidate.generators))
        if len(generators) == target:
            break
    return generators

def get_stabilizer_groups(n_qubits, n_states):
    positive_groups = get_positive_stabilizer_groups(n_qubits, n_states)
    if n_states == n_stabilizer_states(n_qubits):
        print("Found {} positive groups".format(len(positive_groups)))
    groups = [list(map(array_to_pauli, group)) for group in positive_groups]
    sign_strings = get_sign_strings(n_qubits, n_states)
    return add_sign_to_groups(groups, sign_strings)


class _Result(object):
    """Base class to handle pretty-printing of results from a stabilizer rank
    search. Inhertied classes should over-write ostring with a format string
    which takes the arguments:
    target_state: A string description to the target state.
    n_qubits: The number of qubits in the search
    chi: The number returned by the search
    success: Did the method converge/find a smaller stabilizer rank.
    decomposition: A string description of the decomposition found. """
    ostring = None
    def __init__(self, target_state_string, n_qubits, chi,
                 success, decomposition):
        self.target_state = target_state_string
        self.n_qubits = n_qubits
        self.success = success
        self.success_string = 'succeeded' if success else 'did not succeed'
        self.chi = chi
        self.decomposition = decomposition

    def __str__(self):
        if self.ostring is None:
            return """This is the base class. You should not ever be doing this.
                   Why are you doing this. """
        return self.ostring.format(target_state=self.target_state,
                                   n_qubits=self.n_qubits,
                                   success=self.success_string,
                                   chi = self.chi,
                                   decomposition=self.decomposition)


# coding: utf-8

"""The StabilizerSearch.search module contains a pair of base classes,
_Search and _Result, that are used to find and return the Stabilizer Rank
decomposition of a given state.

The other ssubmodules each define a strategy for finding the Stabilizer Rank.

Those methods are imported here to be made available to the rest of the package."""

from .brute_force import BruteForceSearch
from .random_walk import RandomWalkSearch



from cmath import exp as cexp
from math import asin, cos, pi, sin, sqrt
from random import random

import numpy as np

def pyget_su2():
    phi = asin(sqrt(random()))
    psi = random()*2*pi
    chi = random()*2*pi
    return np.matrix([[cexp(1j*psi)*cos(phi), cexp(1j*chi)*sin(phi)],
                      [-1*cexp(-1j*chi)*sin(phi), cexp(-1j*psi)*cos(phi)]],
                     dtype=np.complex_)


"""Module for obtaining Stabilizer States as Numpy arrays. The main method 
provided by this module is `get_stabilizer_states`, which is also available in 
the package namespace.

Potential future plans would allow a more general representation than Numpy 
states. Alternatively, we could look to implement a wrapper function capable of 
transforming between a Numpy matrix and other representations."""

from random import sample

from .eigenstates import py_find_eigenstates
from .py_generators import get_stabilizer_groups as py_get_groups
from .utils import n_stabilizer_states

import os.path as path
import pickle


__all__ = ['get_stabilizer_states']


APP_DIR = path.abspath(__file__)
STATE_STRING = '{}_stabs.pkl'
GROUP_STRING = '{}_groups.pkl'


def try_load(format_string, n_qubits, n_states=None):
    f_string = format_string.format(n_qubits)
    package_path = path.join(APP_DIR, 'data', f_string)
    rel_path = path.join('./', f_string)
    if path.exists(package_path):
        with open(package_path, 'rb') as _f:
            items = pickle.load(_f)
    elif path.exists(rel_path):
        with open(rel_path, 'rb') as _f:
            items = pickle.load(_f)
    else:
        items = None
    if items is not None and n_states != n_stabilizer_states(n_qubits):
        return sample(items, n_states)
    return items


def save_to_pickle(items, format_string, n_qubits):
    with open(path.join('./', format_string.format(n_qubits)), 'wb') as f:
        pickle.dump(items, f)
    return


def get_stabilizer_states(n_qubits, n_states=None, **kwargs):
    """Method for returning a set of stabilizer states. It takes the following 
    arguments:
    Positional:
      n_qubits: The number of qubits our stabilizer states will be built out of.
      n_states (Optional): Number of stabilizer states we require. If not 
      specified, defaults to all Stabilier states.
    Keyword:
      use_cached: Boolean, defaults to True and looks in the package or working 
      dir for serialised states or generators.
      generator_backend: Function which searches for the stabilizer generators
      eigenstate_backend: Function which takes sets of stabilizer generators and
      builds the corresponding eigenstates.
      real_only: Return only real-valued stabilizer states
    """
    use_cached = kwargs.pop('use_cached', True)
    generator_func = kwargs.pop('generator_backend', py_get_groups)
    eigenstate_func = kwargs.pop('eigenstate_backend', py_find_eigenstates)
    real_only = kwargs.pop('real_only', False)
    stabilizer_states = None
    get_all = (n_states == n_stabilizer_states(n_qubits))
    if n_states is None:
        get_all = True
        n_states = n_stabilizer_states(n_qubits)
    if use_cached:
        stabilizer_states = try_load(STATE_STRING, n_qubits, n_states)
        if stabilizer_states is None:
            groups = try_load(GROUP_STRING, n_qubits, n_states)
            if groups is not None:
                if get_all:
                    save_to_pickle(groups, GROUP_STRING, n_qubits)
                stabilizer_states = eigenstate_func(groups, n_states, real_only)
    if stabilizer_states is None:
        generators = generator_func(n_qubits, n_states)
        stabilizer_states = eigenstate_func(generators, real_only)
        if use_cached and get_all:
            save_to_pickle(generators, GROUP_STRING, n_qubits)
            save_to_pickle(stabilizer_states, STATE_STRING, n_qubits)
    return stabilizer_states



from ._result import _Result

def dummy_func(n_qubits, target_state, *args, **kwargs):
    """Dummy method for the search class."""
    print("This should be overridden yo.")
    return True, 0, ['No', 'Result']


class _Search(object):
    """Callable base class for the Stabilizer Decomposition searches. This
    class is overridden by specifying a derived instance of _Result, and a
    function that takes the same arguemnts as search.dummy_func, returning if
    the method succeeded, and the resulting stabilizer rank
    and decomposition."""
    Result_Class = _Result
    func = staticmethod(dummy_func)

    def __init__(self, target_state, target_state_string,
                 n_qubits, *args, **kwargs):
        self.target_state = target_state
        self.target_state_string = target_state_string
        self.n_qubits = n_qubits
        self.f_args = args
        self.f_kwargs = kwargs

    def __call__(self):
        success, chi, decomposition = self.func(self.n_qubits, self.target_state,
                                           *self.f_args, **self.f_kwargs)
        return self.Result_Class(self.target_state_string, self.n_qubits,
                                 chi, success, decomposition)



# This file is part of QuTiP: Quantum Toolbox in Python.
#
#    Copyright (c) 2011 and later, The QuTiP Project.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################
import numpy as np
from qutip.fastsparse import fast_csr_matrix
cimport numpy as np
cimport cython

include "sparse_struct.pxi"

@cython.boundscheck(False)
@cython.wraparound(False)
def coo2fast(object A):
    cdef int nnz = A.nnz
    cdef int nrows = A.shape[0]
    cdef int ncols = A.shape[1]
    cdef complex[::1] data = A.data
    cdef int[::1] rows = A.rows
    cdef int[::1] cols = A.cols
    cdef CSR_Matrix out
    init_CSR(&out, nnz, nrows, ncols)
    
    cdef int i, j, iad, j0
    cdef double complex val
    cdef size_t kk
    # Determine row lengths
    for kk in range(nnz):
        out.indptr[rows[kk]] = out.indptr[rows[kk]] + 1
    # Starting position of rows
    j = 0
    for kk in range(nrows):
        j0 = out.indptr[kk]
        out.indptr[kk] = j
        j += j0
    #Do the data
    for kk in range(nnz):
        i = rows[kk]
        j = cols[kk]
        val = data[kk]
        iad = out.indptr[i]
        out.data[iad] = val
        out.indices[iad] = j
        out.indptr[i] = iad+1
    # Shift back
    for kk in range(nrows,0,-1):
        out.indptr[kk] = out.indptr[kk-1]
    out.indptr[0] = 0
    
    return CSR_to_scipy(&out)
    
    
@cython.boundscheck(False)
@cython.wraparound(False)
def arr_coo2fast(complex[::1] data, int[::1] rows, int[::1] cols, int nrows, int ncols):
    cdef int nnz = data.shape[0]
    cdef CSR_Matrix out
    init_CSR(&out, nnz, nrows, ncols)

    cdef int i, j, iad, j0
    cdef double complex val
    cdef size_t kk
    # Determine row lengths
    for kk in range(nnz):
        out.indptr[rows[kk]] = out.indptr[rows[kk]] + 1
    # Starting position of rows
    j = 0
    for kk in range(nrows):
        j0 = out.indptr[kk]
        out.indptr[kk] = j
        j += j0
    #Do the data
    for kk in range(nnz):
        i = rows[kk]
        j = cols[kk]
        val = data[kk]
        iad = out.indptr[i]
        out.data[iad] = val
        out.indices[iad] = j
        out.indptr[i] = iad+1
    # Shift back
    for kk in range(nrows,0,-1):
        out.indptr[kk] = out.indptr[kk-1]
    out.indptr[0] = 0

    return CSR_to_scipy(&out)
    
    
@cython.boundscheck(False)
@cython.wraparound(False)
def dense2D_to_fastcsr_cmode(complex[:, ::1] mat, int nrows, int ncols):
    cdef int nnz = 0
    cdef size_t ii, jj
    cdef np.ndarray[complex, ndim=1, mode='c'] data = np.zeros(nrows*ncols, dtype=complex)
    cdef np.ndarray[int, ndim=1, mode='c'] ind = np.zeros(nrows*ncols, dtype=np.int32)
    cdef np.ndarray[int, ndim=1, mode='c'] ptr = np.zeros(nrows+1, dtype=np.int32)

    for ii in range(nrows):
        for jj in range(ncols):
            if mat[ii,jj] != 0:
                ind[nnz] = jj
                data[nnz] = mat[ii,jj]
                nnz += 1
        ptr[ii+1] = nnz

    if nnz < (nrows*ncols):
        return fast_csr_matrix((data[:nnz], ind[:nnz], ptr), shape=(nrows,ncols))
    else:
        return fast_csr_matrix((data, ind, ptr), shape=(nrows,ncols))


@cython.boundscheck(False)
@cython.wraparound(False)
def dense2D_to_fastcsr_fmode(complex[::1, :] mat, int nrows, int ncols):
    cdef int nnz = 0
    cdef size_t ii, jj
    cdef np.ndarray[complex, ndim=1, mode='c'] data = np.zeros(nrows*ncols, dtype=complex)
    cdef np.ndarray[int, ndim=1, mode='c'] ind = np.zeros(nrows*ncols, dtype=np.int32)
    cdef np.ndarray[int, ndim=1, mode='c'] ptr = np.zeros(nrows+1, dtype=np.int32)

    for ii in range(nrows):
        for jj in range(ncols):
            if mat[ii,jj] != 0:
                ind[nnz] = jj
                data[nnz] = mat[ii,jj]
                nnz += 1
        ptr[ii+1] = nnz

    if nnz < (nrows*ncols):
        return fast_csr_matrix((data[:nnz], ind[:nnz], ptr), shape=(nrows,ncols))
    else:
        return fast_csr_matrix((data, ind, ptr), shape=(nrows,ncols))




cimport chaar_random
import numpy as np

def get_so2():
    cdef long double **mat = chaar_random.random_so2()
    return np.matrix([[mat[0][0], mat[0][1]],
                      [mat[1][0], mat[1][1]]], dtype=np.complex_)

def get_su2():
    cdef long double complex **mat = chaar_random.random_su2();
    return np.matrix([[mat[0][0], mat[0][1]],
                      [mat[1][0], mat[1][1]]], dtype=np.complex_)



"""Module to hold useful matrices"""

from cmath import exp as cexp
from math import pi, sqrt

from .haar_random import get_so2, get_su2
from .py_haar_random import pyget_su2


import numpy as np


__all__=['qeye', 'X', 'Y', 'Z', 'S', 'H', 'T']


def qeye(n):
    return np.eye(n, dtype=np.complex_)


X = np.matrix([[0,1], [1,0]], dtype=np.complex_)


Y = np.matrix([[0,-1j], [1j,0]], dtype=np.complex_)


Z = np.matrix([[1,0], [0,-1]], dtype=np.complex_)


H = (1/sqrt(2.))*np.matrix([[1, 1],[1, -1]], dtype=np.complex_)


S = np.matrix([[1, 0], [0, 1j]], dtype=np.complex_)


T = np.matrix([[1, 0], [0, cexp(1j*pi/4)]], dtype=np.complex_)



from Cython.Build import cythonize
from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.extension import Extension

import os
import subprocess

def pre_build_dependencies():
    build_root = os.path.dirname(os.path.abspath(__file__))
    os.chdir(os.path.join(build_root, 'stabilizer_search', 'mat'))
    subprocess.call(["make", "clean"])
    subprocess.call(["make", "all"])
    os.chdir(build_root)


class PreBuildInstall(install):

    def run(self):
        pre_build_dependencies()
        install.run(self)


class PreBuildExt(build_ext):

    def run(self):
        pre_build_dependencies()
        build_ext.run(self)

EXTENSIONS = [
    Extension(
        "stabilizer_search.mat.haar_random",
        ["stabilizer_search/mat/haar_random.pyx"],
        include_dirs=['./stabilizer_search/mat/'],
        library_dirs=['./stabilizer_search/mat/'],
        libraries=["m"],
        extra_objects=["./stabilizer_search/mat/haarrandom.a"]
        )
]


setup(
    name='stabilizer_search',
    version='1.0.3',
    url="https://github.com/padraic-padraic/StabilizerSearch",
    author="Padraic Calpin",
    description='Stabilizer Search',
    license="GPLv3.0",
    keywords='quantum computing stabilizers simulation',
    packages=find_packages(exclude=['tests', '*tests']),
    package_data={
        'stabilizer_search.stabilizers':['/data/*.pkl']
    },
    install_requires=['cython', 'numpy', 'scipy', 'six'],
    ext_modules=cythonize(EXTENSIONS),
    cmdclass={'install': PreBuildInstall,
              'build_ext': PreBuildExt}
)



"""Module which defines methods for converting the QObj Pauli Operators that 
generate each stabilizer group, building a projector and finding the associated 
+1 eigenstate."""

from math import pow as fpow
from numpy import allclose, imag
from numpy.linalg import eig

from ..mat import qeye

def find_projector(generating_set):
    n_qubits = len(generating_set)
    _id = qeye(pow(2, n_qubits))
    res = qeye(pow(2, n_qubits))
    for _g in generating_set:
        res = res * (_id+_g)
    return res/fpow(2, n_qubits)

def find_eigenstate(projector):
    eigs, vecs = eig(projector)
    for _n, _eig in enumerate(eigs):
        if allclose(_eig, complex(1)) or allclose(_eig, 1.):
            return (vecs[:,_n])
    return None

def py_find_eigenstates(generating_sets, real_only=False):
    """ """
    states = [find_eigenstate(x) for x in map(find_projector, generating_sets)]
    if real_only:
        return list(filter(lambda x: allclose(imag(x), 0.), states))
    return states
